{% block subject %}{{ subject|default:"Password Reset Request" }}{% endblock %}
{% block body %}
    Hi {{ user.get_short_name }},
    <p>Please use <strong>{{ password }}</strong> as temporary password to reset your {{ app_name }} account password.
    </p>

    <p><strong>NOTE:</strong> the password will expire shortly!</p>

    <p>You received this email because you recently requested for a password reset for your {{ app_name }} account.
        If you did not, kindly discard or ignore this email.</p>
{% endblock %}